#!/usr/bin/python
# this is the driver for predator hunting party
# it runs 4 predators in parallel (general, BFS, DFS/limit1, DFS/limit2)
# during run we test results of all predators each 1 second
# and if decision is reached, we kill all other running predators
# script returns TRUE or FALSE[(classification)] or UNKNOWN
# there is no internal limitation of CPU time or memory

import subprocess
import time
import shlex
import sys
import os
import signal
import argparse
import tempfile
import shutil

predators = None        # list of all 4 predators running in paralllel

witness_dfs_200 = None
witness_dfs_900 = None
witness_bfs = None

def onSIGKILL(signal, frame):
  if predators:
    predators.kill()
  sys.exit(0) # TODO: why 0?

signal.signal(signal.SIGINT, onSIGKILL)
signal.signal(signal.SIGTERM, onSIGKILL)

class PredatorProcess:
  def __init__(self, executable, title, witness=None):
    self.executable = executable
    self.answer = None
    self.title = title
    self.process = None
    self.witness = witness

  def execute(self):
    if self.process is None:
        self.process = subprocess.Popen(self.executable, stdout=subprocess.PIPE, preexec_fn=os.setpgrp)

  def answered(self):
    if not self.terminated():
      return False
    if self.answer is not None:
      return True
    output = self.process.communicate()[0]
    if output is not None:
      self.answer = output.split()[0]
      return True
    return False

  def getAnswer(self):
    if self.answered():
        return (self.answer, self.witness)
    else:
        return (None, self.witness)

  def terminated(self):
    return self.process.poll() is not None

  def kill(self):
    if not self.terminated():
      os.kill(-self.process.pid, signal.SIGKILL)
      self.process.wait()

class PredatorBatch:
  def __init__(self, predator_list):
    self.predators = predator_list

  def launch(self):
    for predator in self.predators:
      predator.execute()

  def kill(self):
    for predator in self.predators:
      predator.kill()
    self.predators = []

  def get_all_answers(self):
    state=[]
    self.running_count=0
    for predator in self.predators[:]:
      if predator.terminated():
         state.append(predator.getAnswer())
      else:
         state.append((None,None))
         self.running_count=self.running_count+1
    return state

  def running(self):
    return self.running_count

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--propertyfile", dest="propertyfile")
  parser.add_argument("--witness", dest="witness")
  parser.add_argument("testcase")
  args = parser.parse_args()

  witness_dfs_200 = tempfile.mkstemp()[1]
  witness_dfs_900 = tempfile.mkstemp()[1]
  witness_bfs = tempfile.mkstemp()[1]

  predator_accelerated = PredatorProcess(shlex.split("predator/sl_build/check-property.sh --trace=/dev/null --propertyfile %s -- %s -m32" % (args.propertyfile, args.testcase)), "Accelerated")
  predator_dfs_200 = PredatorProcess(shlex.split("predator-dfs/sl_build/check-property.sh --trace=/dev/null --propertyfile %s --xmltrace %s --depth 200 -- %s -m32" % (args.propertyfile, witness_dfs_200, args.testcase)), "DFS 200", witness_dfs_200)
  predator_dfs_900 = PredatorProcess(shlex.split("predator-dfs/sl_build/check-property.sh --trace=/dev/null --propertyfile %s --xmltrace %s --depth 900 -- %s -m32" % (args.propertyfile, witness_dfs_900, args.testcase)), "DFS 900", witness_dfs_900)
  predator_bfs = PredatorProcess(shlex.split("predator-bfs/sl_build/check-property.sh --trace=/dev/null --propertyfile %s --xmltrace %s -- %s -m32" % (args.propertyfile, witness_bfs, args.testcase)), "BFS", witness_bfs)

  ## predator_accelerated should be first
  predators = PredatorBatch([predator_accelerated, predator_dfs_200, predator_dfs_900, predator_bfs])

  predators.launch()

  answer = None
  while True:
    state = predators.get_all_answers()          ## list of tuples (answer,witness)
    truecount=0
    falseXFScount=0
    whichXFS=0
    i=0
    for rw in state:
        r = rw[0]
        if r == "TRUE":
            truecount = truecount+1
        if i>0 and (r is not None) and r[0] == "F": ## FALSE check (not for base predator)
            falseXFScount = falseXFScount+1
            whichXFS=i
        i=i+1

    if truecount>0:                             ## any TRUE
        answer="TRUE"
        break

    if falseXFScount>0:                         ## any FALSE* variant
        answer = state[whichXFS][0]
        witness = state[whichXFS][1]
        if witness and args.witness:
           shutil.copyfile(witness, args.witness)
           # unlink postponed to end of script
        break

    if predators.running() == 0:                ## all tasks terminated
        answer="UNKNOWN"
        break

    time.sleep(1)
    # end loop

  predators.kill()                              ## kill all remaining predators if any
  print(answer)
  os.unlink(witness_bfs)
  os.unlink(witness_dfs_200)
  os.unlink(witness_dfs_900)

